package ca.spottedleaf.moonrise.common.misc;

import ca.spottedleaf.concurrentutil.util.IntPairUtil;
import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap;
import it.unimi.dsi.fastutil.longs.LongSet;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMap;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceOpenHashMap;
import it.unimi.dsi.fastutil.objects.ReferenceSet;

public final class PositionCountingAreaMap<T> {

    private final Reference2ReferenceMap<T, PositionCounter> counters = Reference2ReferenceMaps.synchronize(new Reference2ReferenceOpenHashMap<>()); // Canvas - synchronize
    private final Long2IntOpenHashMap positions = new Long2IntOpenHashMap();
    private final Object lock = new Object(); // Canvas - lock

    public ReferenceSet<T> getObjects() {
        return this.counters.keySet();
    }

    public LongSet getPositions() {
        synchronized (lock) { // Canvas - lock
        return this.positions.keySet();
        } // Canvas - lock
    }

    public int getTotalPositions() {
        synchronized (lock) { // Canvas - lock
        return this.positions.size();
        } // Canvas - lock
    }

    public boolean hasObjectsNear(final int toX, final int toZ) {
        synchronized (lock) { // Canvas - lock
        return this.positions.containsKey(IntPairUtil.key(toX, toZ));
        } // Canvas - lock
    }

    public int getObjectsNear(final int toX, final int toZ) {
        synchronized (lock) { // Canvas - lock
        return this.positions.get(IntPairUtil.key(toX, toZ));
        } // Canvas - lock
    }

    public boolean add(final T parameter, final int toX, final int toZ, final int distance) {
        final PositionCounter existing = this.counters.get(parameter);
        if (existing != null) {
            return false;
        }

        final PositionCounter counter = new PositionCounter(parameter);

        this.counters.put(parameter, counter);

        return counter.add(toX, toZ, distance);
    }

    public boolean addOrUpdate(final T parameter, final int toX, final int toZ, final int distance) {
        final PositionCounter existing = this.counters.get(parameter);
        if (existing != null) {
            return existing.update(toX, toZ, distance);
        }

        final PositionCounter counter = new PositionCounter(parameter);

        this.counters.put(parameter, counter);

        return counter.add(toX, toZ, distance);
    }

    public boolean remove(final T parameter) {
        final PositionCounter counter = this.counters.remove(parameter);
        if (counter == null) {
            return false;
        }

        counter.remove();

        return true;
    }

    public boolean update(final T parameter, final int toX, final int toZ, final int distance) {
        final PositionCounter counter = this.counters.get(parameter);
        if (counter == null) {
            return false;
        }

        return counter.update(toX, toZ, distance);
    }

    private final class PositionCounter extends SingleUserAreaMap<T> {

        public PositionCounter(final T parameter) {
            super(parameter);
        }

        @Override
        protected void addCallback(final T parameter, final int toX, final int toZ) {
            synchronized (lock) { // Canvas - lock
            PositionCountingAreaMap.this.positions.addTo(IntPairUtil.key(toX, toZ), 1);
            } // Canvas - lock
        }

        @Override
        protected void removeCallback(final T parameter, final int toX, final int toZ) {
            synchronized (lock) { // Canvas - lock
            final long key = IntPairUtil.key(toX, toZ);
            if (PositionCountingAreaMap.this.positions.addTo(key, -1) == 1) {
                PositionCountingAreaMap.this.positions.remove(key);
            }
            } // Canvas - lock
        }
    }
}
